/**
 *
 * Tämä on:
 * - Esimerkkikoodi, johon ei tarvitse tehdä muokkauksia!
 * - Esimerkki C-ohjelmasta, joka tekee jonkun puolijärkevän asian.
 * - Lisäksi muistettavaa tulevaan elämään: Älä keksi pyörää väärin uudelleen!
 *
 * Inspired by "The danger of naivete" https://blog.codinghorror.com/the-danger-of-naivete/
 *
 * Muistan saman esimerkin ikäni ihan sivulauseesta jollakin omalla kurssilla 2000-luvun alussa.
 * Pistetään nyt hyvä kiertämään, jos vahingossa tätä sivulausetta ei enää tule muilla kursseilla..
 * Sori, jos juuri äsken kuulit saman nykyisellä Ohjelmointi 1 tai 2 -kurssilla :). Sekin mahdollista..
 *
 * Itse demovastaus kulkee eri linjoilla kuin tämä sovellusesimerkki!
 *
 */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <inttypes.h>

#include "vaihtaja.h"

/** täytä taulukko luvuilla 0, 1, 2, ..., montako - 1. */
void luvut(int64_t montako, int64_t *taulukko){
    for( int64_t i = 0; i < montako ; i++ ) {
        taulukko[i] = i;
    }
}

/**
 * Sekoita taulukon alkioiden järjestys omakeksimällä ja simppelillä
 * algoritmilla. Selkeä, eikö vaan? Käydään kaikki alkiot läpi ja
 * vaihdetaan jokainen niistä keskenään satunnaisesti valitun
 * vastinparin kanssa.
 */
void sekoita_yhdella_tavalla(int64_t koko, int64_t *taulukko){
    for(int64_t i = 0; i < koko ; i++){
        int64_t j = rand() % koko;
        vaihda(&taulukko[i], &taulukko[j]);
    }
}

/**
 * Sekoita taulukon alkioiden järjestys Fisherin ja Yatesin vuonna
 * 1938 kuvailemalla tavalla.
 */
void sekoita_toisella_tavalla(int64_t koko, int64_t *taulukko){
    for(int64_t i = 0; i < koko - 1 ; i++){
        int64_t loput = koko - i;
        int64_t j = i + (rand() % loput);
        vaihda(&taulukko[i], &taulukko[j]);
    }
}


/** Tulosta taulukon luvut */
void tulosta(int64_t koko, int64_t *taulukko){
    for( int64_t i = 0; i < koko ; i++ ) {
        printf("%" PRId64 "%s", taulukko[i], ((i < (koko - 1))?" ":"\n"));
    }
}

/** Kirjoita tarinan opetukseen johtavaa narratiivia suomen kielellä. */
void raportoi(int64_t lkm_voittoja, int64_t lkm_osallistujia, int64_t lkm_arvontoja){
    printf("Voitit arvonnassa %" PRId64 " kertaa (%.2f%%).\n",
           lkm_voittoja, (double) lkm_voittoja / lkm_arvontoja * 100 );

    int64_t lkm_oikeasti = lkm_arvontoja / lkm_osallistujia;
    printf("Tilastollisesti sinun piti saada voittaa %" PRId64 " kertaa (%.2f%%).\n",
           lkm_oikeasti, (double) lkm_oikeasti / lkm_arvontoja * 100 );
}

/** Simuloi n arpalipun sekoitus N kertaa kahdella eri tavalla */
void simuloi(int64_t n, int64_t N){
    int64_t *t = malloc(n * sizeof(int64_t));
    int64_t nvoittoja;

    printf("Arpaliput sekoittaa naapurin itseoppinut arvansekoittaja...\n");
    nvoittoja = 0;
    for(int64_t i = 0; i < N; i++){
        luvut(n, t);
        sekoita_yhdella_tavalla(n, t);
        if (t[0] == n-1) {
            nvoittoja++;
        }
    }
    raportoi (nvoittoja, n, N);

    printf("\nArpaliput sekoittaa Ronald Fisherin haamu...\n");
    nvoittoja = 0;
    for(int64_t i = 0; i < N; i++){
        luvut(n, t);
        sekoita_toisella_tavalla(n, t);
        if (t[0] == n-1) {
            nvoittoja++;
        }
    }
    raportoi (nvoittoja, n, N);

    /* Fiksu lukija huomasi, että yllä oli "copy-paste-modify" -koodia,
     * jota tulee välttää oikeasti! Hyvä havainto! Ratkaisu olisi
     * käyttää funktio-osoittimia (englanniksi "function pointer"),
     * joiden syntaksilla ei sekoiteta päätä tässä esimerkissä!
     */

    printf("\nKotiin viemisiksi: Et varmaan itse halua olla epäreilu arvansekoittaja!\n");
    printf("Selvitä, miten hommat toimii, ja käytä tunnettuja ja toimivaksi todettuja algoritmeja!\n");
    printf("Selvitä, mitä käyttämäsi alusta tarjoaa ja että sen shuffle() on testattu kuntoon!\n");
    printf("Pätee kaikkiin muihinkin, jopa yksinkertaisen näköisiin, algoritmeihin!!\n");
    free(t);
}


int main(int argc, char **argv){
    int64_t arpoja = 4;
    int64_t arvontoja = 100000;
    if (argc>1){
        arpoja = atoi(argv[1]);
    }
    if (argc>2){
        arvontoja = atoi(argv[2]);
    }

    printf("\nTervetuloa Suureen Arvontasimulaattoriin\n");
    printf(  "========================================\n\n");

    printf("Osallistut arvontaan %" PRId64
           " kertaa. Arpalippuja on aina %" PRId64
           " kappaletta.\n", arvontoja, arpoja);

    srand(1234);
    simuloi(arpoja, arvontoja);
    return 0;
}


/* Ja loppuun reunahuomautus:
 *
 * Todellisen maailman kaninkolo syvenee kaikissa asioissa - myös
 * tässä.
 *
 * C:n satunnaislukugeneraattoria rand() ei pidä käyttää, jos haluaa
 * olla mahdollisimman reilu tai tietoturvallinen. Oikeassa tieteessä
 * ja VARSINKIN kryptografiassa käytä niihin tarkoituksiin tehtyjä,
 * testattuja ja toimivaksi havaittuja menettelyjä myös ihan
 * satunnaislukujen tuottamisessa!
 *
 * Tähän esimerkkiin se näytti olevan ihan riittävän reilu.
 */
